from clingo_ast_util import *
from pd_utils import get_recursive_rules, get_exit_rules, is_int
import networkx as nx
from networkx.algorithms.components.strongly_connected import strongly_connected_components
import clingo.ast, clingo_ast_util
from clingo.ast import ASTType, ComparisonOperator
import time, random, numpy

positive_label = '+'
negative_label = '-'

class CannotAnalyzeException(BaseException):
    pass

debug = False

class DataHolder():

    def __init__(self, control):
        self.control = control
        self.edb = {} # Only in facts, maps predicate symbol to corresponding calculated number of occurrences
        self.edb_test = {}
        self.idb = set() # All others
        self.selectivity = {} # Maps predicate symbol to corresponding calculated number of occurrences of each argument
        self.min_vals = {}
        self.max_vals = {}
        self.modules = {}
        self.dg = nx.DiGraph()
        self.cg = nx.DiGraph()
        self.acs = None # Admissible component sequence

    def is_IDB(self, predicate):
        '''Returns whether or not a predicate is IDB (not only in facts)'''
        return predicate in self.idb

    def is_EDB(self, predicate):
        '''Returns whether or not a predicate is EDB (only in facts)'''
        return predicate in self.edb
    
    def add_IDB(self, predicate_symbol):
        self.idb.add(predicate_symbol)
    
    def add_EDB(self, predicate, increment=1):
        if predicate not in self.edb:
            self.edb[predicate] = (increment,) * predicate.arity()
        else:
            self.edb[predicate] = tuple(s+increment for s in self.edb[predicate])
    
    def add_EDB_test(self, literal):
        try:
            predicate_symbol = get_predicate_symbol(literal)
        except TypeError:
            raise CannotAnalyzeException()
        args = get_arguments(literal)
        self.add_EDB_by_predicate_args(predicate_symbol, args)
    
    def add_EDB_by_predicate_args(self, predicate_symbol, args):
        stop = False
        for i in range(0,len(args)):
            arg = args[i]
            if arg.type == ASTType.Interval:
                stop = True
                try:
                    l = int(str(arg.left))
                except ValueError:
                    if arg.left.type == ASTType.BinaryOperation:
                        arg.left = self.control.get_const(str(arg.left.left))
                        if arg.left is None:
                            arg.left = self.control.get_const(str(arg.left.right))
                    else:
                        arg.left = self.control.get_const(str(arg.left))
                    if arg.left is None:
                        raise Exception('Failed to parse interval element in: %s' % arg)
                    l = int(str(arg.left))
                try:
                    r = int(str(arg.right))
                except:
                    if arg.right.type == ASTType.BinaryOperation:
                        # this will have to do for now, limited case...
                        arg.right = self.control.get_const(str(arg.right.left))
                        if arg.right is None:
                            arg.right = self.control.get_const(str(arg.right.right))
                    else:
                        arg.right = self.control.get_const(str(arg.right))
                    if arg.right is None:
                        raise Exception('Failed to parse interval element in: %s' % arg)
                    r = int(str(arg.right))
                for interval_val in range(l, r+1):
                    simple_args = list(args)
                    simple_args[i] = clingo.ast.Symbol(arg.location, clingo.Number(interval_val))
                    self.add_EDB_by_predicate_args(predicate_symbol, simple_args)
        if not stop:
            if predicate_symbol not in self.edb_test:
                self.edb_test[predicate_symbol] = tuple(set([ASTWrapper(arg)]) for arg in args)
            else:
                for index, arg in enumerate(args):
                    self.edb_test[predicate_symbol][index].add(ASTWrapper(arg))
    
    def edb_predicates(self):
        return self.edb_test.keys()
    
    def get_edb_statistic(self, predicate):
        if predicate not in self.edb:
            raise KeyError('Non-existant EDB predicate: %s' % predicate)
        return self.edb[predicate]
    
    def increment_selectivity(self, predicate, index, increment=1):
        dprint('increment_selectivity(%s,%s)' % (predicate, index))
        if predicate not in self.selectivity:
            pred_stats = [0] * predicate.arity() # TODO Should this be 1? Or maybe None?
            pred_stats[index] = increment
            self.selectivity[predicate] = pred_stats
        else:
            self.selectivity[predicate][index] += increment

    def get_variable_selectivity(self, predicate, index):
        # if predicate not in self.selectivity:
        #     raise ValueError('Unknown statistics for predicate: %s' % predicate) # TODO Should this actually be 0 to handle when a predicate is in the body of a rule but never in the head (and therefore won't have statistics but should be accounted for as 0)
        # return self.selectivity[predicate][index]
        if predicate not in self.selectivity:
            return 0 # Either something weird happened and we haven't calculated it yet or it doesn't exist in any rule heads
        return self.selectivity[predicate][index]
    
    def reset_selectivity(self):
        '''Resets selectivity to only information provided by EDB predicates'''
        dprint('Resetting to: %s' % self.edb)
        self.selectivity = dict(self.edb)
    
    def reset_selectivity_test(self):
        dprint('Resetting (test) to: %s' % self.edb)
        self.selectivity = {}
        for predicate_symbol in self.edb_test.keys():
            self.selectivity[predicate_symbol] = tuple(len(vals) for vals in self.edb_test[predicate_symbol])
    
    def update_min_max(self):
        for predicate_symbol in self.edb_test.keys():
            stringified_edb_vals = [[str(wv.unwrap()) for wv in self.edb_test[predicate_symbol][arg_index]] for arg_index in range(0, predicate_symbol.arity())]
            dprint(stringified_edb_vals)
            if all((all(is_int(sv) for sv in svl) for svl in stringified_edb_vals)):
                self.min_vals[predicate_symbol] = tuple(min(int(sv) for sv in svl) for svl in stringified_edb_vals)
                self.max_vals[predicate_symbol] = tuple(max(int(sv) for sv in svl) for svl in stringified_edb_vals)
        dprint('MIN VALS: %s' % self.min_vals)
        dprint('MAX VALS: %s' % self.max_vals)
    
    def set_min_val(self, predicate_symbol, arg_index, val):
        self._set_arg_val(self.min_vals, predicate_symbol, arg_index, val)

    def set_max_val(self, predicate_symbol, arg_index, val):
        self._set_arg_val(self.max_vals, predicate_symbol, arg_index, val)

    def _set_arg_val(self, arg_dict, predicate_symbol, arg_index, val):
        if predicate_symbol not in arg_dict:
            arg_dict[predicate_symbol] = tuple((None if ai != arg_index else val) for ai in range(0, predicate_symbol.arity()))
        else:
            # Leave all values but the one we want to change alone
            arg_dict[predicate_symbol] = tuple((arg_dict[predicate_symbol][ai] if ai != arg_index else val) for ai in range(0, predicate_symbol.arity()))

    def get_NDV(self, predicate_symbol, arg_index):
        if predicate_symbol in self.edb_test:
            return len(self.edb_test[predicate_symbol][arg_index])
        else:
            raise Exception('Only EDB predicates are currently supported for NDV (tried %s[%s])' % (predicate_symbol, arg_index))
    
    def get_num_tuples(self, predicate):
        if predicate not in self.selectivity:
            raise ValueError('Unknown statistics for predicate: %s' % predicate)
        num_tuples = 1
        for s in self.selectivity[predicate]:
            num_tuples *= s
        return num_tuples

    def idb_predicates(self):
        return self.idb

    def add_to_module(self, predicate, rule):
        '''Adds a mapping to the module for an individual predicate.
        A module is the set of rules that define a predicate in its head.'''
        if predicate not in self.modules:
            self.modules[predicate] = set()
        self.modules[predicate].add(ASTWrapper(rule))

    def get_module(self, predicates):
        '''Gets the module for a set of predicates, computed by unioning the modules of all members of predicate_set'''
        module = set()
        for predicate in predicates:
            module = module.union(self.modules.get(predicate, set()))
        return [wrapped_r.unwrap() for wrapped_r in module]

class Analyzer(ASTVisitor):
    
    def __init__(self, control, settings):
        self.control = control
        self.settings = settings
        self.data_holder = DataHolder(control)

    def analyze_program(self, program):
        self.identify_IDB_EDB(program)
        self.build_dependency_graph(program)
        self.build_fast_component_graph()
        self.data_holder.acs = nx.topological_sort(self.data_holder.cg)
        # self.build_component_graph()
        # self.data_holder.acs = find_admissible_component_sequence(self.data_holder.cg)
        self.update()

    def identify_IDB_EDB(self, program):
        ConstantDefiner(self.control).visit(program)
        identifier = PredicateIdentifier(self.data_holder)
        identifier.identify_program(program)
    
    def build_dependency_graph(self, program):
        dgb = DepedencyGraphBuilder(self.data_holder)
        dgb.build_graph(program)

    def build_fast_component_graph(self):
        scc = list(tuple(component) for component in strongly_connected_components(self.data_holder.dg)) # Get SCC and make the components immutable (using tuples)
        for component in scc:
            component_dependency_components = set()
            for predicate in component:
                edges = self.data_holder.dg.in_edges(predicate)
                dependency_components = set(self.find_predicate_component(edge[0], scc) for edge in edges if edge[0] not in component) # Removes predicates that are inside of the component since we don't need to loop through recursive rules
                component_dependency_components = component_dependency_components.union(dependency_components)
            for dependency_component in component_dependency_components:
                self.data_holder.cg.add_edge(dependency_component, component)
        isolated_components = set(scc) - set(self.data_holder.cg.nodes()) # Components that don't have any edges to other nodes
        for ic in isolated_components:
            self.data_holder.cg.add_node(ic)
    
    def find_predicate_component(self, predicate, scc):
        '''Finds the corresponding component in the set of SCCs for predicate'''
        for component in scc:
            if predicate in component:
                return component
        raise Exception('Can\'t find component for predicate: %s' % predicate)
    
    def build_component_graph(self):
        scc = set(tuple(component) for component in strongly_connected_components(self.data_holder.dg)) # Get SCC and make the components immutable (using tuples)
        for component_A in scc:
            module = self.data_holder.get_module(component_A)
            for component_B in scc:
                if component_A == component_B:
                    continue
                negative = False
                positive = False
                for rule in module:
                    # We know that a predicate from component_A is in the head already.
                    # Now we just check if a predicate from component_B is in the positive body or negative body.
                    if any(is_predicate_in_positive_body(rule, pred_B) for pred_B in component_B):
                        self.data_holder.cg.add_edge(component_B, component_A, label=positive_label)
                        positive = True
                    elif any(is_predicate_in_negative_body(rule, pred_B) for pred_B in component_B):
                        negative = True
                if negative and not positive:
                    self.data_holder.cg.add_edge(component_B, component_A, label=negative_label)
        isolated_components = scc - set(self.data_holder.cg.nodes()) # Components that don't have any edges to other nodes
        for ic in isolated_components:
            self.data_holder.cg.add_node(ic)

    def update(self):
        # self.data_holder.reset_selectivity()
        self.data_holder.reset_selectivity_test()
        self.data_holder.update_min_max()
        updated_rules = set() # Set of rules that have been updated. This is to prevent double-counting the same rules in two different modules (this can happen with disjunctions)
        for component in self.data_holder.acs:
            dprint('Updating module: %s' % str(component))
            try:
                self.update_module(component, updated_rules)
            except CannotAnalyzeException:
                dprint('WARNING: Found recursive rule. These cannot be analyzed well yet, so statistics for these rules may be underestimates.')

    def update_module(self, component, updated_rules):
        # s = self.data_holder.edb_predicates()
        # ns = set()
        # ds = set()
        module = self.data_holder.get_module(component)
        recursive_rules = get_recursive_rules(component, module)
        exit_rules = get_exit_rules(component, module, recursive_rules=recursive_rules)
        dprint('Recursive rules: %s' % recursive_rules)
        dprint('Exit rules: %s' % exit_rules)
        for rule in exit_rules:
            if ASTWrapper(rule) not in updated_rules:
                self.update_statistics(rule)
                updated_rules.add(ASTWrapper(rule))
        for rule in recursive_rules: # NOTE: Recursive rules are only handled once, so estimates may potentially be lower than the actual values
            if ASTWrapper(rule) not in updated_rules:
                self.update_statistics(rule)
                updated_rules.add(ASTWrapper(rule))
        if len(recursive_rules) > 0: # TODO Remove me (see below)
            raise CannotAnalyzeException('Recursive rules currently not properly supported')
        # continue_loop = True
        # while continue_loop: # TODO Infinite loop here, since we won't ever stop estimating statistics for recursive ones?
            # ds = ns
            # ns = set()
            # s = s.union(ds)
            # for rule in recursive_rules:
            #     self.update_statistics(rule)
            # if len(ns) <= 0:
            #     continue_loop = False
    
    # def update_statistics(self, rule):
    #     dprint('Updating statistics for: %s' % rule)
    #     head_literals = clingo_ast_util.get_head_literals(rule)
    #     for head_literal in head_literals:
    #         dprint('Updating %s' % head_literal)
    #         head_args = get_arguments(head_literal)
    #         w_head_args = [ASTWrapper(ha) for ha in head_args]
    #         head_arg_stats = [1] * len(w_head_args)
    #         head_pred = get_predicate_symbol(head_literal)
    #         # p(A,B,1) :- q(A), r(A,B).
    #         for body_literal in rule.body:
    #             if not is_positive(body_literal): # We can only use positive literals to calculate statistics
    #                 # TODO can we use negative literals to help reduce our overestimations?
    #                 continue
    #             body_args = get_arguments(body_literal)
    #             w_body_args = [ASTWrapper(ba) for ba in body_args]
    #             body_pred = get_predicate_symbol(body_literal)
    #             for head_index in range(0, len(w_head_args)):
    #                 head_arg = w_head_args[head_index]
    #                 if head_arg.unwrap().type != ASTType.Variable:
    #                     continue # We only care about variable arguments in the head
    #                 try:
    #                     body_index = w_body_args.index(head_arg)
    #                 except ValueError:
    #                     continue # This variable is not in this body literal
    #                 s = self.data_holder.get_variable_selectivity(body_pred, body_index)
    #                 head_arg_stats[head_index] = min(head_arg_stats[head_index], s)
    #         head_index = 0
    #         for stat in head_arg_stats:
    #             dprint(head_pred)
    #             dprint(head_arg_stats)
    #             self.data_holder.increment_selectivity(head_pred, head_index, increment=stat)
    #             head_index += 1

    def update_statistics(self, rule): # Update statistics lazily?
        dprint('Updating statistics for: %s' % rule)
        # var_range_sizes = self.compute_nat_join_selectivity(rule.body)
        # if rule.head.type == ASTType.Aggregate:
        #     body_lits = flatten_list([cond_lit.condition for cond_lit in rule.head.elements])
        # else:
        #     body_lits = []
        body_lits = []
        body_lits += rule.body
        selectivity_data = self.compute_nat_join_selectivity_with_min_max(body_lits)
        var_range_sizes = selectivity_data['var_range_sizes']
        dprint('Nat join sel: %s' % var_range_sizes)
        dprint(selectivity_data)
        head_literals = clingo_ast_util.get_head_literals(rule)
        for head_literal in head_literals:
            head_args = get_arguments(head_literal)
            for head_index in range(0, len(head_args)):
                head_arg = head_args[head_index]
                w_head_arg = ASTWrapper(head_arg)
                head_pred = get_predicate_symbol(head_literal)
                # Update min/max info if possible
                if w_head_arg in selectivity_data['var_min']:
                    arg_min = selectivity_data['var_min'][w_head_arg]
                    arg_max = selectivity_data['var_max'][w_head_arg]
                    self.data_holder.set_min_val(head_pred, head_index, arg_min)
                    self.data_holder.set_max_val(head_pred, head_index, arg_max)
                if head_arg.type != ASTType.Variable or w_head_arg not in var_range_sizes: # TODO Handle function args
                    s = 1 # Either not a variable or we don't have any statistic information on this variable (probably in an assignment or something?)
                else:
                    max_s = var_range_sizes.get(w_head_arg) # Maximum possible # of values this variable could be
                    dprint("max s is %s" % max_s)
                    if max_s is None:
                        raise CannotAnalyzeException('Variable %s is not in any positive literal in the body of %s' % (head_arg, rule))
                    s = max_s # TODO Here we try to reduce the size of s based on information we gathered
                self.data_holder.increment_selectivity(head_pred, head_index, increment=s)

    def predict_rule(self, rule):
        selectivity_data = self.compute_nat_join_selectivity_with_min_max(rule.body)
        var_range_sizes = selectivity_data['var_range_sizes']
        var_ignore = selectivity_data['var_ignore']
        var_force = selectivity_data['var_force']
        dprint('Var range sizes: %s' % var_range_sizes)
        rule_size = 1
        for w_var in var_range_sizes.keys():
            if w_var.unwrap().name == '_': # Ignore anonymous variables
                continue
            # elif w_var in var_ignore and w_var not in var_force:
            elif w_var in var_ignore:
                continue
            rule_size *= var_range_sizes[w_var]
        # self.decomp_heuristic_predict(rule)
        return rule_size
    
    def predict_all_rules(self, rule_list, update_statistics=True):
        # TODO Order the rules in rule_list based on component graph so it isn't order dependent
        if type(rule_list) is not list:
            raise TypeError('invalid rule_list type: %s' % type(rule_list))
        rule_list = [rule for rule in rule_list if rule is not None and isinstance(rule, clingo.ast.AST) and rule.type == ASTType.Rule]
        if update_statistics:
            for rule in rule_list:
                if not is_constraint(rule):
                    self.update_statistics(rule)
        return [self.predict_rule(rule) for rule in rule_list if isinstance(rule, clingo.ast.AST) and rule.type == ASTType.Rule]

    def compute_nat_join_size(self, literal_list):
        var_range_sizes = self.compute_nat_join_selectivity(literal_list)
        updated_sizes = {}
        for var in var_range_sizes.keys():
            size = var_range_sizes[var]

            updated_sizes[var] = size
        return updated_sizes

    def compute_nat_join_selectivity_with_min_max(self, literal_list):
        return self._var_selectivity_util_min_max(literal_list, min)

    def compute_nat_join_selectivity(self, literal_list):
        return self._var_selectivity_util_min_max(literal_list, min)['var_range_sizes']

    # def compute_domain_sizes(self, literal_list):
    #     return self._var_selectivity_util(literal_list, max)

    # def _var_selectivity_util(self, literal_list, sel_func):
    #     var_range_sizes = {}
    #     for literal in literal_list:
    #         if not is_positive(literal):
    #             continue
    #         args = get_arguments(literal)
    #         for arg_index in range(0, len(args)):
    #             arg = args[arg_index]
    #             if arg.type != ASTType.Variable: # This will need to change to handle more complex expressions (e.g. X+1)
    #                 continue
    #             w_arg = ASTWrapper(arg)
    #             pred = get_predicate_symbol(literal)
    #             cur_size = var_range_sizes.get(w_arg)
    #             s = self.data_holder.get_variable_selectivity(pred, arg_index)
    #             if cur_size is None:
    #                 var_range_sizes[w_arg] = s # First encounter of this variable
    #             else:
    #                 var_range_sizes[w_arg] = sel_func(s, cur_size)
    #     return var_range_sizes

    def _var_selectivity_util_min_max(self, literal_list, sel_func):
        var_range_sizes = {}
        var_min = {}
        var_max = {}
        var_ignore = set() # Set of variables to ignore for calculating rule sizes
        var_force = set() # Set of variables to use in calculation, even if in the ignore set
        dprint(self.data_holder.min_vals)
        dprint(self.data_holder.max_vals)
        for literal in literal_list:
            if literal.type == ASTType.ConditionalLiteral:
                literal = literal.literal
            if not is_positive(literal):
                continue
                if self.settings['histogram'] and literal.atom.type == ASTType.Comparison:
                    comp_literal = literal.atom
                    self.check_supported_comparator(comp_literal)
                    w_left_vars = find_variables(comp_literal.left)
                    w_left_var = next(iter(w_left_vars)) if len(w_left_vars) > 0 else None
                    w_right_vars = find_variables(comp_literal.right)
                    w_right_var = next(iter(w_right_vars)) if len(w_right_vars) > 0 else None
                    # For now we will assume there is only only 1 variable and 1 symbol
                    # TODO Don't do that
                    # NOTE Only assuming integer values currently
                    w_var = w_left_var if w_left_var is not None else w_right_var
                    val = int(str(comp_literal.left)) if w_left_var is None else int(str(comp_literal.right))
                    if comp_literal.comparison == ComparisonOperator.Equal:
                        # TODO Check if that val is actually in range of stuff, probably just make a flag then check if its in the range later
                        var_min[w_var] = val
                        var_max[w_var] = val
                    elif comp_literal.comparison == ComparisonOperator.GreaterThan:
                        var_min[w_var] = max(var_min[w_var], val + 1)
                    elif comp_literal.comparison == ComparisonOperator.GreaterEqual:
                        var_min[w_var] = max(var_min[w_var], val)
                    elif comp_literal.comparison == ComparisonOperator.LessThan:
                        var_max[w_var] = min(var_max[w_var], val - 1)
                    elif comp_literal.comparison == ComparisonOperator.LessEqual:
                        var_max[w_var] = min(var_max[w_var], val)
                continue
            try:
                args = get_arguments(literal)
            except TypeError:
                args = []
            for arg_index in range(0, len(args)):
                arg = args[arg_index]
                if arg.type == ASTType.Variable:
                    w_var = ASTWrapper(arg)
                elif arg.type == ASTType.UnaryOperation or arg.type == ASTType.BinaryOperation:
                    variables = find_variables(arg)
                    if len(variables) != 1:
                        continue
                    w_var = next(iter(variables))
                else:
                    continue # Handle more as needed
                # if arg.type != ASTType.Variable: # This will need to change to handle more complex expressions (e.g. X+1)
                #     continue
                pred = get_predicate_symbol(literal)
                cur_size = var_range_sizes.get(w_var)
                s = self.data_holder.get_variable_selectivity(pred, arg_index)
                if cur_size is None:
                    var_range_sizes[w_var] = s # First encounter of this variable
                else:
                    var_range_sizes[w_var] = sel_func(s, cur_size)
                if self.settings['histogram']:
                    # Check overlap
                    if pred in self.data_holder.min_vals:
                        if w_var not in var_min or var_min[w_var] is None:
                            var_min[w_var] = self.data_holder.min_vals[pred][arg_index]
                            var_max[w_var] = self.data_holder.max_vals[pred][arg_index]
                        else:
                            # print('var_min[%s] = %s; %s' % (w_var, var_min[w_var], self.data_holder.min_vals[pred][arg_index]))
                            if self.data_holder.min_vals[pred][arg_index] is not None:
                                var_min[w_var] = max(var_min[w_var], self.data_holder.min_vals[pred][arg_index])
                            if self.data_holder.max_vals[pred][arg_index] is not None:
                                var_max[w_var] = min(var_max[w_var], self.data_holder.max_vals[pred][arg_index])
                # Key stuff
                if self.settings['keys']:
                    key = self.settings['keys'].get_key_for_predicate(pred)
                    if key is not None:
                        # Have a set of variables that we ignore and a set that we never ignore (for when they are part of the key sometimes but not always)
                        if key.is_index_keyed(arg_index):
                            var_force.add(w_var)
                        else:
                            var_ignore.add(w_var)
        for var in var_min.keys():
            # TODO For now we are assuming increments of 1 between the range (e.g. a min of 1 and max of 10 will have 10 values in its domain)
            if var_min[var] is None:
                continue
            domain_size = (var_max[var] - var_min[var]) + 1
            if domain_size < 0:
                domain_size = 0
            var_range_sizes[var] = sel_func(var_range_sizes[var], domain_size)
        dprint('Ignore: %s' % var_ignore)
        dprint('Force: %s' % var_force)
        return {'var_range_sizes': var_range_sizes, 'var_min': var_min, 'var_max': var_max, 'var_ignore': var_ignore, 'var_force': var_force}
    
    def check_supported_comparator(self, comp_literal):
        if (comp_literal.left.type != ASTType.Symbol and comp_literal.left.type != ASTType.Variable) or (comp_literal.right.type != ASTType.Symbol and comp_literal.right.type != ASTType.Variable):
            raise TypeError('Comparisons must only contain terms of type Symbol or Variable (found %s, %s)' % (comp_literal.left.type, comp_literal.right.type))
        return True

    # def decomp_heuristic_predict(self, rule):
    #     estimate = 0
    #     test = []
    #     prev_vars = set()
    #     domain_sizes = self.compute_domain_sizes(rule.body)
    #     join_selectivity_values = {}
    #     for body_index, body_literal in enumerate(rule.body): # body_literal is a_(j+1)
    #         if not is_positive(body_literal):
    #             continue
    #         body_pred = get_predicate_symbol(body_literal)
    #         t = self.data_holder.get_num_tuples(body_pred)
    #         body_args = get_arguments(body_literal)
    #         w_body_args = set(ASTWrapper(ba) for ba in body_args)
    #         all_vars_selectivity = 1
    #         nat_join_vars = prev_vars.intersection(w_body_args)
    #         for v in nat_join_vars:
    #             ba_index = body_args.index(v.unwrap())
    #             s = self.data_holder.get_variable_selectivity(body_pred, ba_index)
    #             all_vars_selectivity *= s
    #         left = t / all_vars_selectivity
    #         # Aj_join_selectivity_values = self.compute_nat_join_selectivity(rule.body[:body_index]) # TODO Implement this in here with updates to the dict so we don't run this every time...
    #         right = 1
    #         for v in nat_join_vars:
    #             # right *= Aj_join_selectivity_values[v] / all_join_selectivity_values[v]
    #             right *= join_selectivity_values[v] / domain_sizes[v]
    #         var_args = set(arg for arg in w_body_args if arg.unwrap().type == ASTType.Variable)
    #         prev_vars.update(var_args) # TODO Support function args
    #         test.append(left * right)
    #         for v in var_args:
    #             arg_index = body_args.index(v.unwrap())
    #             s = self.data_holder.get_variable_selectivity(body_pred, arg_index)
    #             if v in join_selectivity_values:
    #                 dprint('Updating %s [%s, %s]' % (v, s, domain_sizes[v]))
    #                 join_selectivity_values[v] = join_selectivity_values[v] * (s / domain_sizes[v])
    #             else:
    #                 join_selectivity_values[v] = s
    #         dprint(join_selectivity_values)
    #     dprint(test)

class ConstantDefiner(ASTVisitor):

    def __init__(self, control):
        self.control = control
    
    def visit_Definition(self, definition):
        with self.control.builder() as b:
            b.add(definition)

class PredicateIdentifier(ASTVisitor):

    def __init__(self, data_holder):
        self.data_holder = data_holder

    def identify_program(self, program):
        for rule in program:
            try:
                self.visit(rule)
            except CannotAnalyzeException:
                continue # Ignore rules we can't analyze
    
    def visit_Rule(self, rule):
        # "It's EDB until proven guilty"
        head_literals = clingo_ast_util.get_head_literals(rule)
        for head_literal in head_literals:
            try:
                predicate_symbol = get_predicate_symbol(head_literal)
            except TypeError:
                raise CannotAnalyzeException()
            if len(rule.body) > 0:
                # Not a fact
                if predicate_symbol in self.data_holder.edb:
                    del self.data_holder.edb[predicate_symbol]
                    del self.data_holder.edb_test[predicate_symbol]
                self.data_holder.add_IDB(predicate_symbol)
                self.data_holder.add_to_module(predicate_symbol, rule)
            elif predicate_symbol not in self.data_holder.idb:
                self.data_holder.add_EDB(predicate_symbol)
                self.data_holder.add_EDB_test(head_literal)

class DepedencyGraphBuilder(ASTVisitor):
    
    def __init__(self, data_holder):
        self.data_holder = data_holder
    
    def build_graph(self, program):
        for rule in program:
            # dprint('Building for %s' % rule)
            try:
                self.visit(rule)
            except CannotAnalyzeException:
                continue # Ignore rules we can't analyze (this should happen for constrains and some complex expressions not supported)
    
    def visit_Rule(self, rule):
        head_literals = clingo_ast_util.get_head_literals(rule)
        for head_literal in head_literals:
            try:
                h_predicate = get_predicate_symbol(head_literal)
            except TypeError as e:
                raise CannotAnalyzeException()
            for body_literal in rule.body:
                try:
                    if is_positive(body_literal):
                        b_predicate = get_predicate_symbol(body_literal)
                    else:
                        continue
                except TypeError:
                    dprint('Warning: skipped abnormal item in rule body while building dependency graph')
                    continue
                if self.data_holder.is_IDB(b_predicate):
                    self.data_holder.dg.add_edge(b_predicate, h_predicate)
                    dprint('Added edge: %s -> %s' % (b_predicate, h_predicate))
                else:
                    self.data_holder.dg.add_node(h_predicate)
                    dprint('Added node: %s' % h_predicate)

def positively_precedes(graph, node_A, node_B):
    for path in nx.all_simple_paths(graph, source=node_A, target=node_B):
        found_negative = False
        for i in range(1, len(path)):
            from_node = path[i-1]
            to_node = path[i]
            label = get_label(graph, from_node, to_node)
            if label != positive_label:
                found_negative = True
                break
        if not found_negative:
            return True
    return False

def negatively_precedes(graph, node_A, node_B):
    for path in nx.all_simple_paths(graph, source=node_A, target=node_B):
        for i in range(1, len(path)):
            from_node = path[i-1]
            to_node = path[i]
            label = get_label(graph, from_node, to_node)
            if label == negative_label:
                return True
    return False
            
def get_label(graph, from_node, to_node):
    return graph[from_node][to_node]['label']

def find_admissible_component_sequence(component_graph):
    # We will build the sequence backwards
    remaining_components = list(component_graph.nodes())
    sequence = []
    prev_len = -1
    while len(remaining_components) > 0:
        if prev_len == len(remaining_components):
            no_change = True # We found all components that both do NOT positively and negatively precede any other remaining components
        else:
            no_change = False
        prev_len = len(remaining_components)
        index_A = 0
        for component_A in remaining_components:
            # print(component_A)
            found_positive_precedents = False
            found_negative_precedents = False
            for component_B in remaining_components:
                if component_A == component_B:
                    continue
                if positively_precedes(component_graph, component_A, component_B):
                    found_positive_precedents = True
                    break # We can't add component_A yet because component_B depends on component_A
                if not no_change and negatively_precedes(component_graph, component_A, component_B):
                    # We only care if we haven't found all components that both do NOT positively and negatively precede
                    found_negative_precedents = True
                    break
            if not found_positive_precedents and (no_change or not found_negative_precedents):
                # Insert it at the front of the sequence because nothing that we insert in front of component_A will depend on it,
                # but items already in the sequence after it may still depend on it. Then, remove it from the remaining components
                # so that we do not consider it anymore (we don't even need to consider it as component_B since everything else will
                # be inserted before it).
                sequence.insert(0, component_A)
                remaining_components.pop(index_A)
            index_A += 1
    return sequence

def dprint(msg):
    if debug:
        print(msg)

def flatten_list(l):
    flat_list = []
    for sublist in l:
        for item in sublist:
            flat_list.append(item)
    return flat_list
